@@ -3,6 +3,9 @@
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
+#ifdef EMCC2
+#include <emscripten.h>
+#endif
 #include "first.h"
 #ifdef KHICAS
 int main(){
@@ -41,7 +44,7 @@ using namespace std;
 #include <fstream>
 #include <iomanip>
 #include <time.h>
-#ifdef HAVE_SYS_TIME_H
+#if defined HAVE_SYS_TIME_H || defined EMCC2
 #include <sys/time.h>
 #endif
 //#include <unistd.h> // For reading arguments from file
@@ -167,8 +170,9 @@ void dealloc(struct Tgraph *graph);
 
 static int texmacs_counter= 0;
 
-
+#ifndef EMCC2
 #include "Xcas1.h"
+#endif
 
 #ifndef HAVE_LIBFLTK
 using namespace giac;
@@ -457,7 +461,7 @@ static const string silent_err(" 2>/dev/null");
 #endif
 
 bool texmacs_graph_lr_margins(const string &fname,int &val,bool init=true,int width=0) {
-#ifdef VISUALC
+#if defined VISUALC || defined USE_GMP_REPLACEMENTS
   return false;
 #else
   char buffer[1024];
@@ -669,9 +673,11 @@ void texmacs_output(const giac::gen & g,giac::gen & gg,bool reading_file,int no,
           std::map<std::string,std::string>::const_iterator it=giac::lexer_localization_map().find(c),itend=giac::lexer_localization_map().end();
           if (it!=itend)
             c=it->second;
+#if !defined EMCC && !defined EMCC2
           vector<string> v=giac::html_help(giac::html_mtt,c);
           if (!v.empty())
             ;
+#endif
         }
         print_help_text(desc,usg,rel,ex);
       }
@@ -696,7 +702,9 @@ void texmacs_output(const giac::gen & g,giac::gen & gg,bool reading_file,int no,
     }
   } else {
     int ans_num=(int)history_out(contextptr).size()-1;
+#ifndef USE_GMP_REPLACEMENTS
     printf("scheme:(equation* (document %s))",giac::gen2scm(gg,giac::context0).c_str());
+#endif
   }
 #else
   if (reading_file){
@@ -1418,8 +1426,43 @@ int micropyjs_evaled(string & s,const giac::context * contextptr){
   return 0;
 }
 
+#ifdef EMCC2
+std::string get_string(const std::string & in){
+  char buf[1024];
+  EM_ASM({
+      var msg=UTF8ToString($0);
+      var jsString = prompt(msg);
+      if (jsString==null) jsString="";
+      var lengthBytes = lengthBytesUTF8(jsString)+1;
+      stringToUTF8(jsString,$1,1024);
+      },in.c_str(),buf);
+  std::string s(buf);
+  //free(buf);
+  return s;
+}
+#endif
+
+void end_icas(){
+#ifdef HAVE_SIGNAL_H
+  if (giac::child_id>1){
+    cerr << "Killing child process " << giac::child_id << "\n";
+    kill(giac::child_id,SIGKILL);
+    usleep(1);
+    if (1){
+      unlink(giac::cas_entree_name().c_str());
+      unlink(giac::cas_sortie_name().c_str());
+    } else {
+      cerr << "Fork files may be present,  rm -f " <<  giac::cas_entree_name() << " " << giac::cas_sortie_name() << "\n";
+    }
+  }
+#endif
+  if (getenv("GIAC_RELEASE")) // for valgrind
+    giac::release_globals();
+}  
+
 int main(int ARGC, char *ARGV[]){
   //giac::stack_check_init(512*1024);  
+  bool inmcp=((ARGC>=2) && std::string(ARGV[1])=="--mcp");
   bool inemacs=((ARGC>=2) && std::string(ARGV[1])=="--emacs");
   bool insage=((ARGC>=2) && std::string(ARGV[1])=="--sage");
   bool intexmacs=((ARGC>=2) && std::string(ARGV[1])=="--texmacs");
@@ -1440,6 +1483,20 @@ int main(int ARGC, char *ARGV[]){
   giac::context * contextptr = 
     //  (giac::context *) giac::context0 ; 
     &giac_context;
+#ifdef EMCC2
+  for (int i=0;;++i){
+    string s;
+    s=get_string("Empty string -> end");
+    if (s.size()==0)
+      break;
+    cout << ">>> " << s << "\n" ;
+    giac::gen g(s,contextptr);
+    g=eval(g,eval_level(contextptr),contextptr);
+    cout << g.print(contextptr) << "\n";
+  }
+  end_icas();
+  return 0;
+#endif  
   bool dohevea=true;
   if (ARGC>1 && strcmp(ARGV[ARGC-1],"--pdf")==0)
     dohevea=false;
@@ -1501,6 +1558,7 @@ int main(int ARGC, char *ARGV[]){
       out << tmp.print(contextptr) << ' ' << char(10) << char(13) ;
       // out << "// Time " << double(end-start)/CLOCKS_PER_SEC << char(10) << char(13)  << char(10) << char(13) ;
     }
+    end_icas();
     return 0;
   }
 #endif
@@ -1508,6 +1566,7 @@ int main(int ARGC, char *ARGV[]){
     cout << "// (c) 2001, 2021 B. Parisse & others" << '\n';
     cout << GIAC_VERSION << '\n';
 #ifndef GNUWINCE
+    end_icas();
     return 0;
 #endif
   }
@@ -1580,11 +1639,13 @@ int main(int ARGC, char *ARGV[]){
       }
       if (!giac::is_file_available(infile.c_str())){
 	COUT << "Unable to read " << infile << '\n';
+        end_icas();
 	return 1;
       }
       pgiac(infile,outfile,checkptr,checkptrin,dohevea);
       if (checkptr) delete checkptr;
       if (checkptrin) delete checkptrin;
+      end_icas();
       return 0;
     }
   }
@@ -1616,17 +1677,21 @@ int main(int ARGC, char *ARGV[]){
     }
     if (!giac::is_file_available(infile.c_str())){
       COUT << "Unable to read " << infile << '\n';
+      end_icas();
       return 1;
     }
     pgiac(infile,outfile,checkptr,checkptrin,dohevea);
     if (checkptr) delete checkptr;
     if (checkptrin) delete checkptrin;
+    end_icas();
     return 0;
   }
   // Help and completion
   int helpitems;
+  // export GIAC_DEBUG==-2 to output static_help*
+  // export GIAC_DEBUG==-3 or -4 to output mcp tool list
 #ifdef STATIC_BUILTIN_LEXER_FUNCTIONS
-  if (giac::debug_infolevel==-2){
+  if (giac::debug_infolevel<=-2 && giac::debug_infolevel>=-4){
 #endif
     giac::readhelp((*giac::vector_aide_ptr()),"aide_cas",helpitems,false);
     if (!helpitems){
@@ -1645,29 +1710,30 @@ int main(int ARGC, char *ARGV[]){
       cerr << "Registered " << helpitems << " commands" << '\n';
   }
   giac::set_language(giac::language(contextptr),contextptr);
-
   /* *******
    * EMACS *
    *********/
-  // #define EMACS_DEBUG 1
+  //#define EMACS_DEBUG 1
 #if !defined EMCC && !defined EMCC2 && !defined NSPIRE_NEWLIB && !defined KHICAS
-  if (inemacs){
+  if (inemacs || inmcp){
     giac::html_help_init(ARGV[0],false);
     int out_handle;
 #ifdef WITH_GNUPLOT
     giac::run_gnuplot(out_handle);
 #endif
-    putchar(EMACS_DATA_END);
-    putchar(EMACS_DATA_BEGIN);
-    putchar(EMACS_RESULT);
-    printf("Giac CAS for mupacs, released under the GPL license 3.0\n");
-    printf("See http://www.gnu.org for license details\n");
-    printf("May contain BSD licensed software parts (lapack, atlas, tinymt)\n");
-    printf("| (c) 2006, 2021 B. Parisse & al (giac), F.Maltey & al (mupacs) |\n");
-    putchar(EMACS_DATA_END);
+    if (inemacs){
+      putchar(EMACS_DATA_END);
+      putchar(EMACS_DATA_BEGIN);
+      putchar(EMACS_RESULT);
+      printf("Giac CAS for mupacs, released under the GPL license 3.0\n");
+      printf("See http://www.gnu.org for license details\n");
+      printf("May contain BSD licensed software parts (lapack, atlas, tinymt)\n");
+      printf("| (c) 2006, 2021 B. Parisse & al (giac), F.Maltey & al (mupacs) |\n");
+      putchar(EMACS_DATA_END);
+    }
     bool prompt=true;
     for (int k=0;;++k) {
-      if (prompt){
+      if (prompt && inemacs){
 	putchar(EMACS_DATA_BEGIN);
 	putchar(EMACS_PROMPT);
 	printf(">> ");
@@ -1678,18 +1744,22 @@ int main(int ARGC, char *ARGV[]){
       string buffer;
       char car;//,nxt;
 #ifdef EMACS_DEBUG
-      ofstream logfile(("log"+giac::print_INT_(k)).c_str());
+      ofstream logfile(("log"+giac::print_INT_(k%16)).c_str());
 #endif
       for (;;){
 	int i=getchar();
 	car=i;
 #ifdef EMACS_DEBUG
-	if (i<32)
-	  logfile << "Ctrl-" << i << '\n';
-	else
-	  logfile << car << '\n';
+        if (i!=-1){
+          if (i<32)
+            logfile << "Ctrl-" << giac::print_INT_(i) << '\n';
+          else {
+            logfile << car << '\n';
+          }
+          logfile.flush();
+        }
 #endif
-	if (i==EOF)
+	if (inemacs && i==EOF)
 	  break;
 	if (i==EMACS_DATA_END){
 	  buffer = buffer + '\n';
@@ -1785,7 +1855,10 @@ int main(int ARGC, char *ARGV[]){
 	  prompt=false;
 	}
 	if (cmd==EMACS_QUESTION){
-	  if (buffer=="quit")
+#ifdef EMACS_DEBUG
+	  logfile << "cmd: " << buffer << "\n" ;
+#endif
+	  if (buffer=="quit" || buffer=="quit;")
 	    break;
 	  if (buffer=="?")
 	    buffer="?giac";
@@ -1795,7 +1868,7 @@ int main(int ARGC, char *ARGV[]){
 	  giac::gen g(buffer,contextptr),gg;
 	  
 #ifdef HAVE_SIGNAL_H_OLD
-	  if (giac::messages_to_print.size()>1){
+	  if (inemacs && giac::messages_to_print.size()>1){
 	    putchar(EMACS_DATA_BEGIN);
 	    putchar(EMACS_ERROR);
 	    printf("%s\n",giac::messages_to_print.c_str());
@@ -1803,12 +1876,16 @@ int main(int ARGC, char *ARGV[]){
 	  }
 #endif
 	  check_browser_help(g);
-	  putchar(EMACS_DATA_BEGIN);
-	  putchar(EMACS_RESULT);
+          if (inemacs){
+            putchar(EMACS_DATA_BEGIN);
+            putchar(EMACS_RESULT);
+          }
 	  // printf("%s\n",giac::messages_to_print.c_str());
 
 	  // GEO SETUP?
 	  if ( (g.type==giac::_SYMB) && (g._SYMBptr->sommet==giac::at_xyztrange) && (g._SYMBptr->feuille.type==giac::_VECT) && (g._SYMBptr->feuille._VECTptr->size()<12) ){
+	    printf("Done\n");
+	    putchar(EMACS_DATA_END); // end answer
 #ifdef HAVE_LIBFLTK
 	    xcas::Xcas_load_graph_setup(contextptr);
 	    Fl::run();
@@ -1816,8 +1893,6 @@ int main(int ARGC, char *ARGV[]){
 #endif
 	    giac::history_in(contextptr).push_back(g);
 	    giac::history_out(contextptr).push_back(gg);
-	    printf("Done\n");
-	    putchar(EMACS_DATA_END); // end answer
 	    continue;
 	  }
 
@@ -1825,23 +1900,41 @@ int main(int ARGC, char *ARGV[]){
 	  std::string filename;
 #ifdef HAVE_LIBFLTK
 	  xcas::icas_eval(g,gg,reading_file,filename,contextptr);
-	  if (reading_file>=1 || graph_output_type(gg))
-	    printf(xcas::fltk_view(g,gg,"",filename,reading_file,contextptr)?"Done\n":"Plot cancelled or unable to plot\n");
+	  if (reading_file>=1 || graph_output_type(gg)){
+            if (inmcp){
+              printf("Close the graphic window before running next command\n");
+              putchar(EMACS_DATA_END);
+              fflush(stdout);
+              xcas::fltk_view(g,gg,"",filename,reading_file,contextptr);
+            }
+            else {
+              printf(xcas::fltk_view(g,gg,"",filename,reading_file,contextptr)?"Done\n":"Plot cancelled or unable to plot\n");
+              putchar(EMACS_DATA_END);
+              fflush(stdout);
+            }
+          }
 	  else {
 	    giac::history_in(contextptr).push_back(g);
 	    giac::history_out(contextptr).push_back(gg);
 	    printf("%s\n",gg.print(contextptr).c_str());
+            putchar(EMACS_DATA_END);
+            fflush(stdout);
 	  }
 #else
           gg=eval(g,1,contextptr);
           giac::history_in(contextptr).push_back(g);
           giac::history_out(contextptr).push_back(gg);
           printf("%s\n",gg.print(contextptr).c_str());
-#endif
 	  putchar(EMACS_DATA_END);
+          fflush(stdout);
+#endif
+#ifdef EMACS_DEBUG
+	  logfile << "evaled: " << gg.print(contextptr) << "\n" ;
+#endif
 	}
       } // end normal command
     } // end while(1)
+    end_icas();
     return 0;
   } // if (inemacs)
   /* END EMACS */
@@ -1852,7 +1945,11 @@ int main(int ARGC, char *ARGV[]){
   if (intexmacs){
     giac::html_help_init(ARGV[0],false);
     giac::enable_texmacs_compatible_latex_export(true);
+#ifdef USE_GMP_REPLACEMENTS
+    texmacs_image_file="casgraph";
+#else
     texmacs_image_file=giac::temp_file_name("casgraph");
+#endif
 #ifdef WITH_GNUPLOT
     int out_handle;
     giac::run_gnuplot(out_handle);
@@ -2000,13 +2097,15 @@ int main(int ARGC, char *ARGV[]){
 #ifdef HAVE_SIGNAL_H_OLD
       giac::messages_to_print="";
 #endif
-#ifdef HAVE_LIBFLTK
       if (buffer=="xcas"){
+#if 0 // def HAVE_LIBFLTK
 	giac::gen ge; std::string filename;
 	xcas::fltk_view(0,ge,"session.xws",filename,5,contextptr);
+#else
+        giac::system_browser_command("xcas.html");        
+#endif
 	continue;
       }
-#endif
       giac::gen g(buffer,contextptr),gg;
 #ifdef HAVE_SIGNAL_H_OLD
       printf("%s\n",giac::messages_to_print.c_str());
@@ -2079,12 +2178,13 @@ int main(int ARGC, char *ARGV[]){
     if (giac::has_gnuplot)
       giac::kill_gnuplot();
 #endif
+    end_icas();
     return 0;
   }
   /* *********************************************************
    *                       END OF TEXMACS                  *
    ********************************************************* */
-#endif // EMCC not defined
+#endif // EMCC/EMCC2/... not defined
 
   if (insage || getenv("GIAC_NO_TIME"))
     show_time=false;
@@ -2104,7 +2204,7 @@ int main(int ARGC, char *ARGV[]){
       cout << python_console() ;
       if (filestream)
         filestream->close();
-      giac::release_globals();
+      end_icas();
       return 0;
     }
   }
@@ -2152,7 +2252,7 @@ int main(int ARGC, char *ARGV[]){
 #endif
 	continue;
       }
-#ifdef HAVE_LIBFLTK
+#if 0 // def HAVE_LIBFLTK
       if (s=="!xcas"){
 	system("./xcas");
 	printf("%s\n","Running ./xcas");
@@ -2163,6 +2263,11 @@ int main(int ARGC, char *ARGV[]){
 	xcas::fltk_view(0,ge,"session.xws",filename,5,contextptr);
 	continue;
       }
+#else
+      if (s=="xcas"){
+        giac::system_browser_command("xcas.html");
+        continue;
+      }
 #endif
       if (s=="giac"){
 	python_compat(python_compat(contextptr)&3,contextptr);
@@ -2302,6 +2407,7 @@ int main(int ARGC, char *ARGV[]){
     if (giac::has_gnuplot)
       giac::kill_gnuplot();
 #endif
+    end_icas();
     return 0;
   }
 #endif // HAVE_LIBREADLINE
@@ -2407,7 +2513,7 @@ int main(int ARGC, char *ARGV[]){
 #ifdef __APPLE__
       startc=clock();
 #endif
-#ifdef HAVE_LIBFLTK
+#ifndef EMCC2
       xcas::icas_eval(gq,e,reading_file,filename,contextptr);
 #else
       e=eval(gq,1,contextptr);
@@ -2466,8 +2572,7 @@ int main(int ARGC, char *ARGV[]){
 #endif
   if (filestream)
     filestream->close();
-  if (getenv("GIAC_RELEASE")) // for valgrind
-    giac::release_globals();
+  end_icas();
   return resultat;
 }
 #endif // KHICAS
